# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
import time

if sys.version_info[0] == 2:
    import Queue as queue
else:
    import queue

import atexit
import weakref

import gevent
import tine

from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
)


class TineCommand(CommandObject):
    def __init__(
        self,
        name,
        command_name,
        tinename=None,
        username=None,
        ListArgs=None,
        timeout=1000,
        **kwargs,
    ):
        CommandObject.__init__(self, name, username, **kwargs)
        self.commandName = command_name
        self.tineName = tinename
        self.timeout = int(timeout)

    def __call__(self, *args, **kwargs):
        self.emit("commandBeginWaitReply", (str(self.name()),))
        if len(args) == 0:
            commandArgument = []
        else:
            commandArgument = args[0]
        try:
            ret = tine.set(
                self.tineName, self.commandName, commandArgument, self.timeout
            )
            self.emit("commandReplyArrived", (ret, str(self.name())))
        except IOError as strerror:
            logging.getLogger("user_level_log").exception("TINE: %s" % strerror)
            self.emit("commandFailed", (strerror))
            raise strerror
        except Exception as ex:
            logging.getLogger("user_level_log").exception("TINE: error: %s" % str(ex))
            self.emit("commandFailed", (str(ex)))
            raise ex

    def get(self):
        result = None
        try:
            result = tine.get(self.tineName, self.commandName, self.timeout)
        except IOError as strerror:
            logging.getLogger("HWR").error("%s" % strerror)
        except Exception:
            logging.getLogger("HWR").exception("")
        return result

    def abort(self):
        pass

    def is_connected(self):
        return True


def emitTineChannelUpdates():
    while not TineChannel.updates.empty():
        try:
            channel_obj_ref, value = TineChannel.updates.get()
        except queue.Empty:
            break
        else:
            channel_object = channel_obj_ref()
            if channel_object is not None:
                try:
                    channel_object.emit("update", (value,))
                except BaseException:
                    logging.getLogger("HWR").exception(
                        "Exception while emitting new value for channel %s",
                        channel_object.name(),
                    )


def do_tine_channel_update(sleep_time):
    while True:
        emitTineChannelUpdates()
        time.sleep(sleep_time)


class TineChannel(ChannelObject):
    attach = {"timer": tine.attach, "event": tine.notify, "datachange": tine.update}

    updates = queue.Queue()
    updates_emitter = gevent.spawn(do_tine_channel_update, 0.1)

    def __init__(
        self, name, attribute_name, tinename=None, username=None, timeout=1000, **kwargs
    ):
        ChannelObject.__init__(self, name, username, **kwargs)

        self.attributeName = attribute_name
        self.tineName = tinename
        self.timeout = int(timeout)
        self.value = None
        self.oldvalue = None

        self.callback_fail_counter = 0

        self.show_callback_error = False
        self.verbose_error_message = None

        logging.getLogger("HWR").debug(
            "Attaching TINE channel: %s %s" % (self.tineName, self.attributeName)
        )

        if kwargs.get("verbose"):
            logging.getLogger("HWR").debug(
                "GUI logging TINE channel: %s %s" % (self.tineName, self.attributeName)
            )
            self.show_callback_error = True
            if kwargs.get("message"):
                self.verbose_error_message = kwargs.get("message")

        if kwargs.get("size"):
            self.linkid = TineChannel.attach[kwargs.get("attach", "timer")](
                self.tineName,
                self.attributeName,
                self.tineEventCallback,
                self.timeout,
                int(kwargs["size"]),
            )
        else:
            self.linkid = TineChannel.attach[kwargs.get("attach", "timer")](
                self.tineName, self.attributeName, self.tineEventCallback, self.timeout
            )

        if self.linkid > 0 and kwargs.get("attach", "timer") == "datachange":
            tolerance = kwargs.get("tolerance", 0.0)
            try:
                tine.tolerance(self.linkid, 0.0, float(tolerance.rstrip("%")))
            except AttributeError:
                if tolerance != 0.0:
                    tine.tolerance(self.linkid, float(tolerance), 0.0)

        atexit.register(self.__del__)

    def __del__(self):
        try:
            tine.detach(self.linkid)
            logging.getLogger("HWR").debug(
                "TINE channel %s %s detached" % (self.tineName, self.attributeName)
            )
            self.linkid = -1
        except IOError as strerror:
            logging.getLogger("HWR").error(
                "%s detaching %s %s" % (strerror, self.tineName, self.attributeName)
            )
        except Exception:
            logging.getLogger("HWR").error(
                "Exception on detaching %s %s" % (self.tineName, self.attributeName)
            )

    def tineEventCallback(self, id, cc, data_list):
        if cc == 0:
            self.callback_fail_counter = 0
            self.update(data_list)
        elif (
            cc != 103
        ):  #  was str(cc) # and self.attributeName not in ("dozor-pass", "ff-ssim"):
            self.callback_fail_counter = self.callback_fail_counter + 1
            if self.show_callback_error:
                logging.getLogger("GUI").error(
                    "Tine event callback error %s, Channel: %s, Server: %s/%s"
                    % (str(cc), self.name(), self.tineName, self.attributeName)
                )
                if self.verbose_error_message is not None:
                    logging.getLogger("GUI").error("%s" % self.verbose_error_message)

            else:
                logging.getLogger("HWR").error(
                    "Tine event callback error %s, Channel: %s, Server: %s/%s"
                    % (str(cc), self.name(), self.tineName, self.attributeName)
                )

            """
            if self.callback_fail_counter >= 3:
                logging.getLogger("HWR").error(
                    "Repeated tine event callback errors %s, Channel: %s, Server: %s/%s"
                    % (str(cc), self.name(), self.tineName, self.attributeName)
                )
            """

    def update(self, value=None):
        if value is None:
            logging.getLogger("HWR").warning(
                "Update with value None on: %s %s" % (self.tineName, self.attributeName)
            )
            value = self.get_value()
        self.value = value

        if value != self.oldvalue:
            TineChannel.updates.put((weakref.ref(self), value))
            self.oldvalue = value

    def get_value(self, force=False):
        # GB: if forced while having a value already, i.e. well after connecting a channel, do a real synchronous get and return
        if force:
            if self.value is not None:
                logging.getLogger("HWR").warning(
                    "Executing synch get on: %s %s"
                    % (self.tineName, self.attributeName)
                )
                return self._synchronous_get()
            else:
                logging.getLogger("HWR").warning(
                    "Attempting to force unconnected channel: %s %s"
                    % (self.tineName, self.attributeName)
                )
                return None

        # GB: if there is no value yet, wait and hope it will appear somehow:
        _counter = 10
        while self.value is None and _counter <= 100:
            time.sleep(0.020)
            logging.getLogger("HWR").warning(
                "No update after %d ms on: %s %s, executing synchronous get"
                % (_counter * 20, self.tineName, self.attributeName)
            )
            # but now tine lib should be standing the get, so we try....

            self.value = self._synchronous_get()
            _counter += 1
        if self.value is None:
            logging.getLogger("HWR").warning(
                "Gave up waiting for a first update on: %s %s"
                % (self.tineName, self.attributeName)
            )
        return self.value

    def _synchronous_get(self):
        try:
            value = tine.get(self.tineName, self.attributeName, self.timeout)
            return value
        except IOError as strerror:
            logging.getLogger("HWR").error("%s" % strerror)

    def set_value(self, newValue):
        listData = newValue
        try:
            ret = tine.set(self.tineName, self.attributeName, listData, self.timeout)
        except IOError as strerror:
            logging.getLogger("HWR").error("%s" % strerror)

    def is_connected(self):
        return self.linkid > 0

    def set_old_value(self, oldValue):
        self.oldvalue = oldValue



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
Qt4_TangoLimaVideo

[Description]
HwObj used to grab images via Tango Lima device server
If you want to access the Lima Library directly you may consider using
the Qt4_LimaVideo module instead

[Configuration]
Example Hardware Object XML file :
==================================
<object class="Qt4_LimaVideo">
   <type>basler</type>
   <encoding>yuv422p</encoding>
   <tangoname>bl13/eh/lima_oav</tangoname>
   <gain>0.5</gain>
   <exposure>0.01</exposure>
   <mirror>(False, False)</mirror>
   <interval>30</interval>
</object>
"""

import logging

import gevent
from camera import camera
from GenericVideoDevice import GenericVideoDevice

from mxcubecore.utils.qt_import import (
    QImage,
    QPixmap,
)


class PX2Video(GenericVideoDevice, camera):
    """
    Descript. :
    """

    def __init__(self, name):
        """
        Descript. :
        """
        GenericVideoDevice.__init__(self, name)
        camera.__init__(self)
        self.log = logging.getLogger("user_level_log")
        self.device = name
        self.camera = camera()
        self.width = 1360
        self.height = 1024

    def init(self):
        """
        Descript. :
        """
        # tangoname = self.get_property("tangoname")

        self.log.info("PX2Video init")

        self.device = self.prosilica

        GenericVideoDevice.init(self)

    def set_cam_encoding(self, cam_encoding):
        if cam_encoding == "yuv422p":
            self.device.video_mode = "YUV422"
        elif cam_encoding == "y8":
            self.device.video_mode = "Y8"

        GenericVideoDevice.set_cam_encoding(self, cam_encoding)

    def get_video_live(self):
        return True

    def set_video_live(self, flag):
        return
        self.device.video_live = flag

    def get_image_dimensions(self):
        return self.camera.get_image_dimensions()

    def get_image(self):
        return self.camera.get_image(), self.width, self.height

    def get_new_image(self):
        """
        Descript. :
        """
        raw_buffer, width, height = self.get_image()

        if raw_buffer is not None:
            qimage = QImage(raw_buffer, width, height, QImage.Format_RGB888)

            qpixmap = QPixmap(qimage)
            self.emit("imageReceived", qpixmap)
            return qimage

    # def get_jpg_image(self):
    # image = self.camera.get_image()

    # self.emit("imageReceived", qpixmap)
    # return qimage

    def do_image_polling(self, sleep_time=1):
        """
        Descript. :
        """
        while self.get_video_live() == True:
            self.get_new_image()
            gevent.sleep(1)

    """ END Overloading of GenericVideoDevice methods """


def test_hwo(hwo):
    print("Image dimensions: ", hwo.get_image_dimensions())
    print("Live Mode: ", hwo.get_video_live())



import logging

from tango import DeviceProxy

from mxcubecore import HardwareRepository as HWR
from mxcubecore.utils.units import kev_to_ev

log = logging.getLogger("user_level_log")


class PrepareOpenHutch:
    """
    Prepare beamline for opening the hutch door.

    - close safety shutter
    - close detector cover
    - close fast shutter
    - move detector to a safe position
    - put MD3 into 'Transfer' phase in case of OSC delivery mode
    - move MD3's `BeamstopPosition` and `CapillaryPosition` to `PARK` in case of HVE
    - if jungfrau is used, take pedestal
    """

    def __call__(self):
        try:
            # Ensure laser is stopped before opening the hutch
            laser = HWR.beamline.get_object_by_role("laser")
            laser.disarm()

            collect = HWR.beamline.collect
            diffractometer = HWR.beamline.diffractometer
            detector = HWR.beamline.detector

            log.info("Preparing experimental hutch for door opening.")

            collect.close_fast_shutter()
            collect.close_safety_shutter()
            collect.close_detector_cover()

            diffractometer.wait_device_ready()
            if HWR.beamline.is_hve_sample_delivery():
                # This is 'equivalent' of Transfer phase for HVE experiments
                log.info("Setting diffractometer to 'equivalent' of Transfer phase.")
                diffractometer.channel_dict["BeamstopPosition"].set_value("PARK")
                diffractometer.channel_dict["CapillaryPosition"].set_value("PARK")
            else:
                log.info("Setting diffractometer to Transfer phase.")
                diffractometer.set_phase("Transfer")

            log.info("Moving detector to safe position.")
            collect.move_detector_to_safe_position()

            if detector.get_property("model") == "JUNGFRAU":
                log.info("Collecting Jungfrau pedestal.")
                detector.pedestal()

        except Exception as ex:
            # Explicitly add raised exception into the log message,
            # so that it is shown to the user in the beamline action UI log.
            log.exception("Error preparing to open hutch.\nError was: '%s'", str(ex))  # noqa: TRY401


class CheckBeam:
    def __call__(self):
        """
        Check beam stability
        """
        xbpms = {
            "DM3": DeviceProxy("b312a-o06/dia/xbpm-01"),
            "DM4": DeviceProxy("b312a-e01/dia/xbpm-01"),
            "BCU XBPM1": DeviceProxy("b312a-e04/dia/xbpm-01"),
            "BCU XBPM2": DeviceProxy("b312a-e04/dia/xbpm-02"),
        }
        energy = kev_to_ev(HWR.beamline.energy.get_value())
        transmission = HWR.beamline.transmission.get_value()
        for name, xbpm in xbpms.items():
            total_current = xbpm.S
            flux = total_current * (
                -0.534515 * energy**4
                - 43197.6 * energy**3
                + 5.13449e09 * energy**2
                - 4.39169e13 * energy
                + 1.14591e17
            )
            log.info(
                f"XBPM: {name}, total current: {total_current * 1e6:.2f} uA, "
                f"estimated flux at sample position: {flux:.2e} ph/s"
            )

            if "BCU" in name:
                full_flux = flux * 100.0 / transmission
                log.info(
                    f"Current transmission: {transmission:.2f}%, "
                    f"estimated full flux at sample position: {full_flux:.2e} ph/s"
                )


class MeasureFlux:
    def __call__(self):
        """
        calculate flux at sample position
        """
        flux_at_sample = HWR.beamline.collect.get_instant_flux()
        log.info("Flux at sample position is %.2e ph/s", flux_at_sample)


class SaveMD3Position:
    def __call__(self):
        HWR.beamline.diffractometer.save_centered_position()


class MoveToMD3SavedPosition:
    def __call__(self):
        HWR.beamline.diffractometer.goto_centered_position()


class EmptyMount:
    def __call__(self):
        isara = HWR.beamline.sample_changer

        log.info("Performing empty mount recovery sequence.")

        isara.execute_command("Reset")

        log.info("Recovery sequence completed.")



"""A simple client for MetadataManager and MetaExperiment"""

import logging
import math
import os
import smtplib
import sys
import time
import traceback
from email.mime.text import MIMEText

import PyTango.client

from mxcubecore import HardwareRepository as HWR
from mxcubecore.utils.conversion import string_types


class MetadataManagerClient(object):
    """
    A client for the MetadataManager and MetaExperiment tango Devices

    Attributes:
        name: name of the tango device. Example: 'id21/metadata/ingest'
    """

    def __init__(self, metadataManagerName, metaExperimentName):
        """
        Return a MetadataManagerClient object whose metadataManagerName is *metadataManagerName*
        and metaExperimentName is *metaExperimentName*
        """
        self.dataRoot = None
        self.proposal = None
        self.sample = None
        self.datasetName = None

        if metadataManagerName:
            self.metadataManagerName = metadataManagerName
        if metaExperimentName:
            self.metaExperimentName = metaExperimentName

        print("MetadataManager: %s" % metadataManagerName)
        print("MetaExperiment: %s" % metaExperimentName)

        # Tango Devices instances
        try:
            MetadataManagerClient.metadataManager = PyTango.client.Device(
                self.metadataManagerName
            )
            MetadataManagerClient.metaExperiment = PyTango.client.Device(
                self.metaExperimentName
            )
        except Exception:
            print("Unexpexted error: ", sys.exc_info()[0])
            raise

    def printStatus(self):
        print("DataRoot: %s" % MetadataManagerClient.metaExperiment.dataRoot)
        print("Proposal: %s" % MetadataManagerClient.metaExperiment.proposal)
        print("Sample: %s" % MetadataManagerClient.metaExperiment.sample)
        print("Dataset: %s" % MetadataManagerClient.metadataManager.scanName)

    def __setAttribute(self, proxy, attributeName, newValue):
        """
        This method sets an attribute on either the MetadataManager or MetaExperiment server.
        The method checks that the attribute has been set, and repeats up to five times
        setting the attribute if not. If the attribute is not set after five trials the method
        raises an 'RuntimeError' exception.
        """
        caughtException = "unknown exception"
        currentValue = "unknown"
        if newValue == currentValue:
            currentValue = "Current value not known"
        counter = 0
        while counter < 10:
            counter += 1
            try:
                setattr(proxy, attributeName, newValue)
                time.sleep(0.1)
                currentValue = getattr(proxy, attributeName)
                if currentValue == newValue:
                    break
            except Exception as e:
                caughtException = e
                print(
                    "Unexpected error in MetadataManagerClient._setAttribute: {0}".format(
                        e
                    )
                )
                print(
                    "proxy = '{0}', attributeName = '{1}', newValue = '{2}'".format(
                        proxy, attributeName, newValue
                    )
                )
                print("Trying again, trial #{0}".format(counter))
                time.sleep(1)
        if currentValue != newValue:
            raise RuntimeError(
                "Cannot set '{0}' attribute '{1}' to '{2}' due to {3}!".format(
                    proxy, attributeName, newValue, caughtException
                )
            )

    def __setDataRoot(self, dataRoot):
        try:
            MetadataManagerClient.metaExperiment.dataRoot = dataRoot
            self.dataRoot = dataRoot
        except Exception:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    def __setProposal(self, proposal):
        """Set proposal should be done before setting the data root"""
        try:
            MetadataManagerClient.metaExperiment.proposal = proposal
            self.proposal = proposal
        except Exception:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    def appendFile(self, filePath):
        try:
            MetadataManagerClient.metadataManager.lastDataFile = filePath
        except Exception:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    def __setSample(self, sample):
        try:
            MetadataManagerClient.metaExperiment.sample = sample
            self.sample = sample
        except Exception:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    def __setDataset(self, datasetName):
        try:
            MetadataManagerClient.metadataManager.scanName = datasetName
            self.datasetName = datasetName
        except Exception:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    def start(self, dataRoot, proposal, sampleName, datasetName):
        """Starts a new dataset"""
        if MetadataManagerClient.metaExperiment:
            try:
                # setting proposal
                # self.__setProposal(proposal)
                self.__setAttribute(
                    MetadataManagerClient.metaExperiment, "proposal", proposal
                )

                # setting dataRoot
                # self.__setDataRoot(dataRoot)
                self.__setAttribute(
                    MetadataManagerClient.metaExperiment, "dataRoot", dataRoot
                )

                # setting sample
                # self.__setSample(sampleName)
                self.__setAttribute(
                    MetadataManagerClient.metaExperiment, "sample", sampleName
                )

                # setting dataset
                # self.__setDataset(datasetName)
                self.__setAttribute(
                    MetadataManagerClient.metadataManager, "scanName", datasetName
                )
                self.datasetName = datasetName

                # setting datasetName
                if str(MetadataManagerClient.metaExperiment.state()) == "ON":
                    if str(MetadataManagerClient.metadataManager.state()) == "ON":
                        MetadataManagerClient.metadataManager.StartScan()

            except Exception:
                print("Unexpected error:", sys.exc_info()[0])
                raise

    def end(self):
        try:
            MetadataManagerClient.metadataManager.endScan()
        except Exception:
            print("Unexpected error:", sys.exc_info()[0])
            raise

    def get_state(self):
        return str(MetadataManagerClient.metadataManager.state())


class MXCuBEMetadataClient(object):
    def __init__(self, esrf_multi_collect):
        self.esrf_multi_collect = esrf_multi_collect
        if hasattr(self.esrf_multi_collect["metadata"], "manager"):
            self._metadataManagerName = self.esrf_multi_collect["metadata"].manager
        else:
            self._metadataManagerName = None
        if hasattr(self.esrf_multi_collect["metadata"], "experiment"):
            self._metaExperimentName = self.esrf_multi_collect["metadata"].experiment
        else:
            self._metaExperimentName = None
        if hasattr(self.esrf_multi_collect["metadata"], "error_email"):
            self._listEmailReceivers = self.esrf_multi_collect[
                "metadata"
            ].error_email.split(" ")
        else:
            self._listEmailReceivers = []
        if hasattr(self.esrf_multi_collect["metadata"], "replyto_email"):
            self._emailReplyTo = self.esrf_multi_collect["metadata"].replyto_email
        else:
            self._emailReplyTo = None

        self._beamline = HWR.beamline.session.endstation_name

    def reportStackTrace(self):
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        errorMessage = "{0} {1}".format(exc_type, exc_value)
        errorMessage += "\n\n"
        listTrace = traceback.extract_tb(exc_traceback)
        errorMessage += "Traceback (most recent call last): \n"
        for listLine in listTrace:
            errorMessage += '  File "%s", line %d, in %s%s\n' % (
                listLine[0],
                listLine[1],
                listLine[2],
                os.linesep,
            )
        if len(self._listEmailReceivers) > 0 and self._emailReplyTo is not None:
            COMMASPACE = ", "
            mime_text_message = MIMEText(errorMessage)
            replyTo = self._emailReplyTo
            listTo = self._listEmailReceivers
            listCC = []
            listBCC = []
            mime_text_message["Subject"] = (
                "Metadata upload error on {0} for proposal {1}".format(
                    self._beamline, self._proposal
                )
            )
            mime_text_message["From"] = replyTo
            mime_text_message["To"] = COMMASPACE.join(listTo)
            if len(listCC) > 0:
                mime_text_message["CC"] = COMMASPACE.join(listCC)
            if len(listBCC) > 0:
                mime_text_message["BCC"] = COMMASPACE.join(listBCC)
            try:
                smtp = smtplib.SMTP("localhost")
                smtp.sendmail(
                    replyTo, listTo + listCC + listBCC, mime_text_message.as_string()
                )
                smtp.quit()
            except Exception:
                logging.getLogger("HWR").exception("")
        return errorMessage

    def start(self, data_collect_parameters):
        # Metadata
        if (
            self._metadataManagerName is not None
            and self._metaExperimentName is not None
        ):
            try:
                self._proposal = HWR.beamline.session.get_proposal()
                # Create proxy object
                self._metadataManagerClient = MetadataManagerClient(
                    self._metadataManagerName, self._metaExperimentName
                )

                # First check the state of the device server
                serverState = self._metadataManagerClient.get_state()
                if serverState == "RUNNING":
                    # Force end of scan
                    self._metadataManagerClient.end()

                fileinfo = data_collect_parameters["fileinfo"]
                directory = fileinfo["directory"]
                prefix = fileinfo["prefix"]
                run_number = int(fileinfo["run_number"])
                # Connect to ICAT metadata database
                # Strip the prefix from any workflow expTypePrefix
                # TODO: use the ISPyB sample name instead
                sampleName = prefix
                for expTypePrefix in ["line-", "mesh-", "ref-", "burn-", "ref-kappa-"]:
                    if sampleName.startswith(expTypePrefix):
                        sampleName = sampleName.replace(expTypePrefix, "")
                        break
                # The data set name must be unique so we use the ISPyB data collection
                # id
                datasetName = "{0}_{1}_{2}".format(
                    prefix, run_number, self.esrf_multi_collect.collection_id
                )
                self._metadataManagerClient.start(
                    directory, self._proposal, sampleName, datasetName
                )
                self._metadataManagerClient.printStatus()
            except Exception:
                logging.getLogger("user_level_log").warning(
                    "Cannot connect to metadata server"
                )
                errorMessage = self.reportStackTrace()
                logging.getLogger("user_level_log").warning(errorMessage)
                self._metadataManagerClient = None

    def upload_images_to_icat(
        self,
        template,
        prefix,
        run_number,
        directory,
        number_of_images,
        start_image_number,
        overlap,
    ):
        logging.getLogger("user_level_log").info("Uploading to images to ICAT")
        if template.endswith(".h5"):
            if math.fabs(overlap) > 1:
                for image_number in range(1, number_of_images + 1):
                    h5_master_file_name = (
                        "{prefix}_{run_number}_{image_number}_master.h5".format(
                            prefix=prefix,
                            run_number=run_number,
                            image_number=image_number,
                        )
                    )
                    h5_master_file_path = os.path.join(directory, h5_master_file_name)
                    self._metadataManagerClient.appendFile(h5_master_file_path)
                    h5_data_file_name = (
                        "{prefix}_{run_number}_{image_number}_data_000001.h5".format(
                            prefix=prefix,
                            run_number=run_number,
                            image_number=image_number,
                        )
                    )
                    h5_data_file_path = os.path.join(directory, h5_data_file_name)
                    self._metadataManagerClient.appendFile(h5_data_file_path)
            else:
                h5_master_file_name = (
                    "{prefix}_{run_number}_{start_image_number}_master.h5".format(
                        prefix=prefix,
                        run_number=run_number,
                        start_image_number=start_image_number,
                    )
                )
                h5_master_file_path = os.path.join(directory, h5_master_file_name)
                self._metadataManagerClient.appendFile(h5_master_file_path)
                for index in range(int((number_of_images - 1) / 100) + 1):
                    h5_data_file_name = "{prefix}_{run_number}_{start_image_number}_data_{data_index:06d}.h5".format(
                        prefix=prefix,
                        run_number=run_number,
                        start_image_number=start_image_number,
                        data_index=(index + 1),
                    )
                    h5_data_file_path = os.path.join(directory, h5_data_file_name)
                    self._metadataManagerClient.appendFile(h5_data_file_name)
        else:
            for index in range(number_of_images):
                image_no = index + start_image_number
                image_path = os.path.join(directory, template % image_no)
                self._metadataManagerClient.appendFile(image_path)

    def end(self, data_collect_parameters):
        try:
            dictMetadata = self.getMetadata(data_collect_parameters)
            if self._metadataManagerClient is not None:
                # Upload all images
                fileinfo = data_collect_parameters["fileinfo"]
                prefix = fileinfo["prefix"]
                template = fileinfo["template"]
                directory = fileinfo["directory"]
                run_number = int(fileinfo["run_number"])
                for oscillation_parameters in data_collect_parameters[
                    "oscillation_sequence"
                ]:
                    number_of_images = oscillation_parameters["number_of_images"]
                    start_image_number = oscillation_parameters["start_image_number"]
                    overlap = oscillation_parameters["overlap"]
                    self.upload_images_to_icat(
                        template,
                        prefix,
                        run_number,
                        directory,
                        number_of_images,
                        start_image_number,
                        overlap,
                    )
                # Upload the two paths to the meta data HDF5 files
                pathToHdf5File1 = os.path.join(
                    directory,
                    "{proposal}-{beamline}-{prefix}_{run_number}_{dataCollectionId}.h5".format(
                        beamline=self._beamline,
                        proposal=self._proposal,
                        prefix=prefix,
                        run_number=run_number,
                        dataCollectionId=self.esrf_multi_collect.collection_id,
                    ),
                )
                self._metadataManagerClient.appendFile(pathToHdf5File1)
                pathToHdf5File2 = os.path.join(
                    directory,
                    "{proposal}-{prefix}-{prefix}_{run_number}_{dataCollectionId}.h5".format(
                        proposal=self._proposal,
                        prefix=prefix,
                        run_number=run_number,
                        dataCollectionId=self.esrf_multi_collect.collection_id,
                    ),
                )
                self._metadataManagerClient.appendFile(pathToHdf5File2)
                # Upload meta data as attributes
                # These attributes are common for all ESRF MX beamlines
                dictMetadata = self.getMetadata(data_collect_parameters)
                # import pprint
                # pprint.pprint(dictMetadata)
                for attributeName, value in dictMetadata.items():
                    logging.getLogger("HWR").info(
                        "Setting metadata client attribute '{0}' to '{1}'".format(
                            attributeName, value
                        )
                    )
                    setattr(
                        self._metadataManagerClient.metadataManager,
                        attributeName,
                        str(value),
                    )
                self._metadataManagerClient.printStatus()
                self._metadataManagerClient.end()
        except Exception:
            logging.getLogger("user_level_log").warning("Cannot upload metadata")
            errorMessage = self.reportStackTrace()
            logging.getLogger("user_level_log").warning(errorMessage)
            self._metadataManagerClient = None

    def getMetadata(self, data_collect_parameters):
        """
        Common metadata parameters for ESRF MX beamlines.
        """
        listAttributes = [
            ["MX_beamShape", "beamShape"],
            ["MX_beamSizeAtSampleX", "beamSizeAtSampleX"],
            ["MX_beamSizeAtSampleY", "beamSizeAtSampleY"],
            ["MX_dataCollectionId", "collection_id"],
            ["MX_directory", "fileinfo.directory"],
            ["MX_exposureTime", "oscillation_sequence.exposure_time"],
            ["MX_flux", "flux"],
            ["MX_fluxEnd", "flux_end"],
            ["MX_numberOfImages", "oscillation_sequence.number_of_images"],
            ["MX_oscillationRange", "oscillation_sequence.range"],
            ["MX_oscillationStart", "oscillation_sequence.start"],
            ["MX_oscillationOverlap", "oscillation_sequence.overlap"],
            ["MX_resolution", "resolution"],
            ["MX_startImageNumber", "oscillation_sequence.start_image_number"],
            ["MX_scanType", "experiment_type"],
            ["MX_template", "fileinfo.template"],
            ["MX_transmission", "transmission"],
            ["MX_xBeam", "xBeam"],
            ["MX_yBeam", "yBeam"],
            ["InstrumentMonochromator_wavelength", "wavelength"],
        ]
        dictMetadata = {}
        for attribute in listAttributes:
            if isinstance(attribute, list):
                attributeName = attribute[0]
                keyName = attribute[1]
            else:
                attributeName = attribute
                keyName = attribute
            value = None
            if "." in keyName:
                parent, child = str(keyName).split(".")
                parentObject = data_collect_parameters[parent]
                if isinstance(parentObject, type([])):
                    parentObject = parentObject[0]
                if child in parentObject:
                    value = str(parentObject[child])
            elif keyName in data_collect_parameters:
                value = str(data_collect_parameters[keyName])
            if value is not None:
                dictMetadata[attributeName] = value
        # Template - replace python formatting with hashes
        dictMetadata["MX_template"] = dictMetadata["MX_template"].replace(
            "%04d", "####"
        )
        # Motor positions
        motorNames = ""
        motorPositions = ""
        for motor, position in data_collect_parameters["motors"].items():
            if isinstance(motor, string_types):
                motorName = motor
            else:
                nameAttribute = getattr(motor, "name")
                if isinstance(nameAttribute, string_types):
                    motorName = nameAttribute
                else:
                    motorName = nameAttribute()
            if motorNames == "":
                motorNames = motorName
                motorPositions = str(round(position, 3))
            else:
                motorNames += " " + motorName
                if position is not None:
                    motorPositions += " " + str(round(position, 3))
                else:
                    motorPositions += " None"
        dictMetadata["MX_motors_name"] = motorNames
        dictMetadata["MX_motors_value"] = motorPositions
        # Detector distance
        distance = HWR.beamline.detector.distance.get_value()
        if distance is not None:
            dictMetadata["MX_detectorDistance"] = distance
        # Aperture
        if HWR.beamline.beam is not None and HWR.beamline.beam.aperture is not None:
            dictMetadata["MX_aperture"] = HWR.beamline.beam.aperture.get_value()
        return dictMetadata


if __name__ == "__main__":
    metadataManagerName = "id30a1/metadata/ingest"
    metaExperimentName = "id30a1/metadata/experiment"
    client = MetadataManagerClient(metadataManagerName, metaExperimentName)

    client.start(
        "/data/visitor/mx415/id30a1/20161014/RAW_DATA",
        "mx415",
        "sample1",
        "dataset_20161014_1",
    )
    client.appendFile("/data/visitor/mx415/id30a1/20161014/RAW_DATA/t1/test1.txt")
    client.appendFile("/data/visitor/mx415/id30a1/20161014/RAW_DATA/t1/test2.txt")
    client.printStatus()
    client.end()



"""
[Name] PlateManipulator

[Description]
Plate manipulator hardware object is used to use diffractometer in plate mode.
It is compatible with md2, md3 diffractometers. Class is based on
SampleChanger, so it has all the sample changed functionalities, like
mount, unmount sample (in this case move to plate position).
Plate is organized in rows and columns. Each cell (Cell) contains drop (Drop).
Each drop could contain several crystals (Xtal). If CRIMS is available then
each drop could have several crystals.

[Channels]

 - self.chan_current_phase   : diffractometer phase
 - self.chan_plate_location  : plate location (col, row)
 - self.chan_drop_location   : plate location (col, row, drop)
 - self.chan_state           : diffractometer state

[Commands]

 - self.cmd_move_to_location : move to plate location

[Emitted signals]

 - emitted signals defined in SampleChanger class

[Included Hardware Objects]
-----------------------------------------------------------------------
| name            | signals          | functions
-----------------------------------------------------------------------
-----------------------------------------------------------------------
"""

import logging
import time

import gevent

from mxcubecore.HardwareObjects.abstract.AbstractSampleChanger import (
    SampleChanger,
    SampleChangerMode,
    SampleChangerState,
)
from mxcubecore.HardwareObjects.abstract.sample_changer import Crims
from mxcubecore.HardwareObjects.abstract.sample_changer.Container import (
    Basket,
    Container,
    Sample,
)


class Xtal(Sample):
    __NAME_PROPERTY__ = "Name"
    __LOGIN_PROPERTY__ = "Login"

    def __init__(self, drop, index):
        super(Xtal, self).__init__(drop, Xtal._get_xtal_address(drop, index), False)
        self._drop = drop
        self._index = index
        self._set_image_x(None)
        self._set_image_y(None)
        self._set_image_url(None)
        self._set_name(None)
        self._set_login(None)
        self._set_info_url(None)

        self._set_info(False, False, False)
        self._set_loaded(False, False)
        self.present = True

    def _set_name(self, value):
        self._set_property(self.__NAME_PROPERTY__, value)

    def get_name(self):
        return self.get_property(self.__NAME_PROPERTY__)

    def _set_login(self, value):
        self._set_property(self.__LOGIN_PROPERTY__, value)

    def get_login(self):
        return self.get_property(self.__LOGIN_PROPERTY__)

    def get_drop(self):
        return self._drop

    def get_cell(self):
        return self.get_drop().get_cell()

    def get_basket_no(self):
        """
        In this cas we assume a drop is a basket or puck
        """
        return self.get_drop().get_index() + 1

    def get_cell_no(self):
        """
        In this cas we assume wells in the row is a cell
        """
        return self.get_cell().get_row_index() + 1

    @staticmethod
    def _get_xtal_address(drop, index):
        return str(drop.get_address()) + "-" + str(index)

    def get_index(self):
        """
        Descript. : Sample index is calculated relative to the row (Basket)
                    In this case we assume that in drop is one xtal
                    This should be changed to various num of xtals in the drop
        """
        cell_index = self.get_cell().get_index()
        drops_in_cell_num = self.get_cell().get_drops_no()
        drop_index = self._drop.get_index()
        return cell_index * drops_in_cell_num + drop_index

    def get_container(self):
        return self.get_cell().get_container()

    def get_name(self):
        return "%s%d:%d" % (
            self.get_cell().get_row_chr(),
            self.get_cell().get_index() + 1,
            self._drop.get_index() + 1,
        )


class Drop(Container):
    __TYPE__ = "Drop"

    def __init__(self, cell, drops_num):
        super(Drop, self).__init__(
            self.__TYPE__, cell, Drop._get_drop_address(cell, drops_num), False
        )
        self._cell = cell
        self._drops_num = drops_num

    @staticmethod
    def _get_drop_address(cell, drop_num):
        return str(cell.get_address()) + ":" + str(drop_num)

    def get_cell(self):
        return self._cell

    def get_well_no(self):
        return self.get_index() + 1

    def is_loaded(self):
        """
        Returns if the sample is currently loaded for data collection
        :rtype: bool
        """
        sample = self.get_sample()
        return sample.is_loaded()

    def get_sample(self):
        """
        In this cas we assume that there is one crystal per drop
        """
        sample = self.get_components()
        return sample[0]


class Cell(Container):
    __TYPE__ = "Cell"

    def __init__(self, row, row_chr, col_index, drops_num):
        Container.__init__(
            self, self.__TYPE__, row, Cell._get_cell_address(row_chr, col_index), False
        )
        self._row = row
        self._row_chr = row_chr
        self._col_index = col_index
        self._drops_num = drops_num
        for drop_index in range(self._drops_num):
            drop = Drop(self, drop_index + 1)
            self._add_component(drop)
            xtal = Xtal(drop, drop.get_number_of_components())
            drop._add_component(xtal)
        self._transient = True

    def get_row(self):
        return self._row

    def get_row_chr(self):
        return self._row_chr

    def get_row_index(self):
        return ord(self._row_chr.upper()) - ord("A")

    def get_col(self):
        return self._col_index

    def get_drops_no(self):
        return self._drops_num

    @staticmethod
    def _get_cell_address(row, col):
        return str(row) + str(col)


class PlateManipulator(SampleChanger):
    """ """

    __TYPE__ = "PlateManipulator"

    def __init__(self, *args, **kwargs):
        super(PlateManipulator, self).__init__(self.__TYPE__, False, *args, **kwargs)

        self.plate_label = None
        self.num_cols = None
        self.num_rows = None
        self.num_drops = None
        self.current_phase = None
        self.reference_pos_x = None
        self.timeout = 3  # default timeout
        self.plate_location = None
        self.crims_url = None
        self.crims_user_agent = None
        self.plate_barcode = None
        self.harvester_key = None
        self.processing_plan = None

        self.stored_pos_x = None
        self.stored_pos_y = None

        self.cmd_move_to_drop = None
        self.cmd_move_to_location = None

        self.chan_state = None

    def init(self):
        """
        Descript. :
        """
        cmd_get_config = self.get_channel_object("GetPlateConfig", optional=True)
        if cmd_get_config:
            try:
                (
                    self.num_rows,
                    self.num_cols,
                    self.num_drops,
                ) = cmd_get_config.get_value()
            except Exception:
                self.log.exception("")
        else:
            self.num_cols = self.get_property("numCols")
            self.num_rows = self.get_property("numRows")
            self.num_drops = self.get_property("numDrops")

        self.reference_pos_x = self.get_property("referencePosX")
        if not self.reference_pos_x:
            self.reference_pos_x = 0.5

        self.stored_pos_x = self.reference_pos_x
        self.stored_pos_y = 0.5
        self.plate_label = self.get_property("plateLabel")
        self.crims_url = self.get_property("crimsWsRoot")
        self.crims_user_agent = self.get_property("crimsUserAgent")
        self.plate_barcode = self.get_property("PlateBarcode")
        self.harvester_key = self.get_property("harvesterKey")

        self.cmd_move_to_drop = self.get_command_object("MoveToDrop")
        if not self.cmd_move_to_drop:
            self.cmd_move_to_location = self.get_command_object(
                "startMovePlateToLocation"
            )

        self.cmd_move_to_crystal_position = self.get_command_object(
            "MoveToXtalPointing"
        )
        self.cmd_get_omega_scan_limits = self.get_command_object(
            "getOmegaMotorDynamicScanLimits"
        )

        self.cmd_do_abort = self.get_command_object("AbortCurrentAction")

        self._init_sc_contents()

        self.chan_current_phase = self.get_channel_object("CurrentPhase")
        self.chan_drop_location = self.get_channel_object("DropLocation")
        self.chan_plate_location = self.get_channel_object("PlateLocation")
        if self.chan_plate_location is not None:
            self.chan_plate_location.connect_signal(
                "update", self.plate_location_changed
            )
            self.plate_location_changed(self.chan_plate_location.get_value())

        self.chan_state = self.get_channel_object("State")
        if self.chan_state is not None:
            self.chan_state.connect_signal("update", self.state_changed)

        SampleChanger.init(self)

    def change_plate_barcode(self, barcode):
        if self._load_data(barcode):
            self.plate_barcode = barcode
            return True
        else:
            raise Exception("barcode unknown")

    def hw_get_loaded_sample_location(self):
        loaded_sample = None
        if self.chan_drop_location:
            loaded_sample = self.chan_drop_location.get_value()

            return (
                chr(65 + loaded_sample[0])
                + str(loaded_sample[1] + 1)
                + ":"
                + str(loaded_sample[2] + 1)
                + "-0"
            )
        return loaded_sample

    def plate_location_changed(self, plate_location):
        self.plate_location = plate_location
        self._update_loaded_sample()
        self.update_info()

    def state_changed(self, state):
        try:
            self.plate_location_changed(self.chan_plate_location.get_value())
            self._on_state_changed(state)
        except AttributeError:
            self.log.exception("")

    def _on_state_changed(self, state):
        """
        Descript. : state change callback. Based on diffractometer state
                    sets PlateManipulator state.
        """
        if state is None:
            self._set_state(SampleChangerState.Unknown)
        else:
            if state == "Alarm":
                self._set_state(SampleChangerState.Alarm)
            elif state == "Fault":
                self._set_state(SampleChangerState.Fault)
            elif state == "Moving" or state == "Running":
                self._set_state(SampleChangerState.Moving)
            elif state == "Ready":
                if self.current_phase == "Transfer":
                    self._set_state(SampleChangerState.Charging)
                elif self.current_phase == "Centring":
                    self._set_state(SampleChangerState.Ready)
                else:
                    self._set_state(SampleChangerState.StandBy)
            elif state == "Initializing":
                self._set_state(SampleChangerState.Initializing)

    def _init_sc_contents(self):
        """
        Descript. : Initializes content of plate.
        """
        if self.num_rows is None:
            return
        self._set_info(False, None, False)
        self._clear_components()
        for row in range(self.num_rows):
            # row is like a basket
            basket = Basket(self, row + 1, samples_num=0, name="Row")
            present = True
            datamatrix = ""
            scanned = False
            basket._set_info(present, datamatrix, scanned)
            self._add_component(basket)

            for col in range(self.num_cols):
                cell = Cell(basket, chr(65 + row), col + 1, self.num_drops)
                basket._add_component(cell)

    def _do_abort(self):
        """
        Descript. :
        """
        self.cmd_do_abort()

    def _do_change_mode(self, mode):
        """
        Descript. :
        """
        if mode == SampleChangerMode.Charging:
            self._set_phase("Transfer")
        elif mode == SampleChangerMode.Normal:
            self._set_phase("Centring")

    def _do_load(self, sample=None):
        """
        Descript. :
        """
        selected = self.get_selected_sample()
        if sample is None:
            sample = self.get_selected_sample()
        if sample is not None:
            if sample != selected:
                self._do_select(sample)
            self._set_loaded_sample(sample)

    def load(self, sample=None, wait=True):
        comp = self._resolve_component(sample)
        coords = comp.get_coords()
        res = self._load_sample(coords)
        if res:
            self._set_loaded_sample(comp)
            comp._set_loaded(True, True)
        return res

    def _load_sample(self, sample_location=None, pos_x=None, pos_y=None, wait=True):
        """
        Location is estimated by sample location and reference positions.
        """
        try:
            row = sample_location[0] - 1
            col = int((sample_location[1] - 1) / self.num_drops)
            drop = sample_location[1] - self.num_drops * col

            if not pos_x:
                pos_x = self.stored_pos_x
            else:
                self.stored_pos_x = pos_x
            if not pos_y:
                pos_y = self.stored_pos_y
            else:
                self.stored_pos_y = pos_y
                # pos_y = float(drop) / (self.num_drops + 1)

            if self.cmd_move_to_location:
                self.cmd_move_to_location(row, col, pos_x, pos_y)
                if wait:
                    self._wait_ready(60)
            elif self.cmd_move_to_drop:
                self.cmd_move_to_drop(row, col, drop - 1)
                if wait:
                    self._wait_ready(60)
            else:
                # No actual move cmd defined. Act like a mockup
                self.plate_location = [row, col, self.stored_pos_x, pos_y]
                col += 1
                cell = self.get_component_by_address("%s%d" % (chr(65 + row), col))
                drop = cell.get_component_by_address(
                    "%s%d:%d" % (chr(65 + row), col, drop)
                )
                new_sample = drop.get_sample()
                old_sample = self.get_loaded_sample()
                new_sample = drop.get_sample()
                if old_sample != new_sample:
                    if old_sample is not None:
                        old_sample._set_loaded(False, True)
                    if new_sample is not None:
                        new_sample._set_loaded(True, True)

            return True
        except:
            return False

    def _do_unload(self, sample_slot=None):
        """
        Descript. :
        """
        self._reset_loaded_sample()
        self._on_state_changed("Ready")

    def _do_reset(self):
        """
        Descript. :
        """
        self._reset(False)
        self._wait_device_ready()

    def _do_scan(self, component, recursive):
        """
        Descript. :
        """
        if not isinstance(component, PlateManipulator):
            raise Exception("Not supported")
        self._initializeData()
        if self.get_token() is None:
            raise Exception("No plate barcode defined")
        self._load_data(self.get_token())

    def sync_with_crims(self):
        """
        Descript. :
        Get Crims information
        """
        self.processing_plan = self._load_data(self.plate_barcode)
        return self.processing_plan

    def _do_select(self, component):
        """
        Descript. :
        """
        pos_x = self.stored_pos_x
        pos_y = 0.5

        if isinstance(component, Xtal):
            self._select_sample(
                component.get_cell().get_row_index(),
                component.get_cell().get_col() - 1,
                component.get_drop().get_well_no() - 1,
            )
            self._set_selected_sample(component)
            component.get_container()._set_selected(True)
            component.get_container().get_container()._set_selected(True)
        elif isinstance(component, Crims.CrimsXtal):
            col = component.Column - 1
            row = ord(component.Row.upper()) - ord("A")
            pos_x = component.offsetX
            pos_y = component.offsetY
            cell = self.get_component_by_address(
                Cell._get_cell_address(component.Row, component.Column)
            )
            drop = self.get_component_by_address(
                Drop._get_drop_address(cell, component.Shelf)
            )
            drop._set_selected(True)
            drop.get_container()._set_selected(True)
        elif isinstance(component, Drop):
            self._select_sample(
                component.get_cell().get_row_index(),
                component.get_cell().get_col() - 1,
                component.get_well_no() - 1,
            )
            component._set_selected(True)
            component.get_container().get_container()._set_selected(True)
        elif isinstance(component, Cell):
            self._select_sample(component.get_row_index(), component.get_col() - 1, 0)
            component._set_selected(True)
        elif isinstance(component, list):
            row = component[0]
            col = component[1]
            if len(component > 2):
                pos_x = component[2]
                pos_y = component[3]
            cell = self.get_component_by_address(Cell._get_cell_address(row, col))
            cell._set_selected(True)
        else:
            raise Exception("Invalid selection")
        self._reset_loaded_sample()
        self._wait_device_ready()

    def _load_data(self, plate_barcode):
        processing_plan = Crims.get_processing_plan(
            plate_barcode, self.crims_url, self.crims_user_agent, self.harvester_key
        )
        if processing_plan is None:
            msg = (
                "No information about plate with barcode %s found in CRIMS"
                % plate_barcode
            )
            logging.getLogger("user_level_log").error(msg)
        else:
            msg = (
                "Information about plate with barcode %s found in CRIMS" % plate_barcode
            )
            logging.getLogger("user_level_log").info(msg)
            self._set_info(True, processing_plan.plate.barcode, True)

            for x in processing_plan.plate.xtal_list:
                cell = self.get_component_by_address(
                    Cell._get_cell_address(x.row, x.column)
                )
                cell._set_info(True, "", True)
                drop = self.get_component_by_address(
                    Drop._get_drop_address(cell, x.shelf)
                )
                drop._set_info(True, "", True)
                xtal = Xtal(drop, drop.get_number_of_components())
                xtal._set_info(True, x.pin_id, True)
                xtal._set_image_url(x.image_url)
                xtal._set_image_x(x.offset_x)
                xtal._set_image_y(x.offset_y)
                xtal._set_login(x.login)
                xtal._set_name(x.sample)
                xtal._set_info_url(x.summary_url)
                drop._add_component(xtal)
        return processing_plan

    def _do_update_info(self):
        """
        Descript. :
        """
        self._update_state()

    def _read_state(self):
        return self.chan_state.get_value()

    def _update_state(self):
        """
        Descript. :
        """
        state = None
        if self.chan_state is not None:
            state = self.chan_state.get_value()
            if (state == "Ready") or (self.current_phase is None):
                self.current_phase = self.chan_current_phase.get_value()
            self._on_state_changed(state)
        return state

    def _update_loaded_sample(self):
        """Updates plate location"""

        if self.plate_location is not None:
            new_sample = self.get_loaded_sample()
            if new_sample is not None:
                # self._update_sample_barcode(new_sample)
                loaded = True
                has_been_loaded = True
                new_sample._set_loaded(loaded, has_been_loaded)
                self._set_loaded_sample(new_sample)

    def get_loaded_sample(self):
        sample = None
        for s in self.get_sample_list():
            if s.get_address() == self.hw_get_loaded_sample_location():
                sample = s
        return sample

    def get_sample(self, plate_location):
        row = int(plate_location[0])
        col = int(plate_location[1])
        y_pos = float(plate_location[3])
        drop_index = abs(y_pos * self.num_drops) + 1
        if drop_index > self.num_drops:
            drop_index = self.num_drops

        cell = self.get_component_by_address("%s%d" % (chr(65 + row), col + 1))
        if cell:
            drop = cell.get_component_by_address(
                "%s%d:%d" % (chr(65 + row), col + 1, drop_index)
            )
            return drop.get_sample()

    def get_sample_list(self):
        """
        Descript. : This is ugly
        """
        sample_list = []
        for basket in self.get_components():
            if isinstance(basket, Basket):
                for cell in basket.get_components():
                    if isinstance(cell, Cell):
                        for drop in cell.get_components():
                            sample_list.append(drop.get_sample())
        return sample_list

    def is_mounted_sample(self, sample_location):
        row = sample_location[0] - 1
        col = (sample_location[1] - 1) / self.num_drops
        drop = sample_location[1] - self.num_drops * col
        pos_y = float(drop) / (self.num_drops + 1)
        sample = self.get_sample((row, col, drop, pos_y))
        return sample.loaded

    def _ready(self):
        if self._update_state() == "Ready":
            return True
        return False

    def _wait_ready(self, timeout=None):
        if timeout <= 0:
            timeout = self.timeout
        tt1 = time.time()
        while time.time() - tt1 < timeout:
            if self._ready():
                break
            else:
                gevent.sleep(0.5)

    def get_plate_info(self):
        """
        Descript. : returns dict with plate info
        """
        plate_info_dict = {}
        plate_info_dict["plate_label"] = self.plate_label or "Demo plate label"
        plate_info_dict["plate_barcode"] = self.plate_barcode or ""
        plate_info_dict["num_cols"] = self.num_cols
        plate_info_dict["num_rows"] = self.num_rows
        plate_info_dict["num_drops"] = self.num_drops
        return plate_info_dict

    def get_plate_location(self):
        if self.chan_plate_location is not None:
            self.plate_location = self.chan_plate_location.get_value()
        return self.plate_location

    def move_to_crystal_position(self, crystal_uuid):
        """
        Descript. : Move Diff to crystal position
        Get crystal_uuid from processing plan for loaded sample/drop
        """
        ret = None
        if crystal_uuid in ["undefined", None]:
            loaded_sample = self.chan_drop_location.get_value()
            row = int(loaded_sample[0])
            col = int(loaded_sample[1])
            drop = int(loaded_sample[2])

            if self.processing_plan:
                for x in self.processing_plan.plate.xtal_list:
                    if (
                        row == ord(x.row) - 65
                        and col == x.column - 1
                        and drop == x.shelf - 1
                    ):
                        crystal_uuid = x.crystal_uuid
            else:
                raise Exception("No processing_plan OR Crystal Found in this well")

        if self.cmd_move_to_crystal_position and crystal_uuid:
            try:
                # ret = self.cmd_move_to_crystal_position(row, col, drop, x.image_url, x.offset_x, x.offset_y, 0.0, 0.0, False)
                ret = self.cmd_move_to_crystal_position(
                    self.plate_barcode, crystal_uuid
                )
            except Exception as ex:
                raise Exception("Could not move to crystal position %s" % str(ex))
        else:
            raise Exception(
                "move_to_crystal_position command or crystal UUID not found"
            )

        return ret

    def get_scan_limits(self, args):
        """
        get Omega Motor Dynamic Scan Limits
        """
        ret = ""
        if self.cmd_get_omega_scan_limits:
            try:
                ret = self.cmd_get_omega_scan_limits(args)
            except Exception:
                raise Exception("Could not get Omega Motor Dynamic Scan Limits")
        else:
            raise Exception("command not found")

        return ret



import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.mxcubecore.HardwareObjects.ProposalTypeISPyBLims import (
    ProposalISPyBClient,
)


class ALBAISPyBClient(ProposalISPyBClient):
    def ldap_login(self, login_name, psd, ldap_connection):
        # overwrites standard ldap login  is ISPyBClient.py
        #  to query for homeDirectory

        if ldap_connection is None:
            ldap_connection = self.ldapConnection

        ok, msg = ldap_connection.login(
            login_name, psd, fields=["uid", "homeDirectory"]
        )

        if ok:
            vals = ldap_connection.get_field_values()
            if "homeDirectory" in vals:
                home_dir = vals["homeDirectory"][0]
                self.log.debug(
                    "  homeDirectory for user %s is %s" % (login_name, home_dir)
                )
                # HWR.beamline.session.set_base_data_directories(home_dir, home_dir, home_dir)
                HWR.beamline.session.set_ldap_homedir(home_dir)
        else:
            home_dir = "/tmp"
            HWR.beamline.session.set_ldap_homedir(home_dir)

        return ok, msg

    def translate(self, code, what):
        """
        Given a proposal code, returns the correct code to use in the GUI,
        or what to send to LDAP, user office database, or the ISPyB database.
        """
        self.log.debug("translating %s %s" % (code, what))
        if what == "ldap":
            if code == "mx":
                return "u"

        if what == "ispyb":
            if code == "u":
                return "mx"

        return code

    def prepare_collect_for_lims(self, mx_collect_dict):
        # Attention! directory passed by reference. modified in place

        for i in range(4):
            try:
                prop = "xtalSnapshotFullPath%d" % (i + 1)
                path = mx_collect_dict[prop]
                ispyb_path = HWR.beamline.session.path_to_ispyb(path)
                logging.debug("ALBA ISPyBClient - %s is %s " % (prop, ispyb_path))
                mx_collect_dict[prop] = ispyb_path
            except Exception:
                self.log.exception("")

    def prepare_image_for_lims(self, image_dict):
        for prop in ["jpegThumbnailFileFullPath", "jpegFileFullPath"]:
            try:
                path = image_dict[prop]
                ispyb_path = HWR.beamline.session.path_to_ispyb(path)
                image_dict[prop] = ispyb_path
            except Exception:
                self.log.exception("")


def test_hwo(hwo):
    proposal = "mx2018012551"
    proposal = "mx2018002222"
    pasw = "2222008102"

    info = hwo.login(proposal, pasw)
    # info = hwo.get_proposal(proposal_code, proposal_number)
    # info = hwo.get_proposal_by_username("u2020000007")
    print(info["status"])

    print("Getting associated samples")
    session_id = 58248
    proposal_id = 8250
    samples = hwo.get_samples(proposal_id, session_id)
    print(samples)



import logging
import os
import pickle
import pwd
import socket
import time

import gevent
import gevent.server

from mxcubecore.BaseHardwareObjects import Procedure
from mxcubecore.utils import qt_import

"""
<procedure class="InstanceServer">
  <host>myhostname</host>
  <port>myport</port>
</procedure>
"""

TERMINATOR = "\0"
INSTANCE_HO = None
SERVER_CLIENTS = {}
CLIENTS = {}

###
# Creates an asynchronous TCP server to manage multiple application instances
###


class QtInstanceServer(Procedure):
    # Initializes the hardware object
    def init(self):
        # Read the HO configuration
        self.serverPort = self.get_property("port")
        self.serverHost = self.get_property("host")
        if self.serverHost is None:
            self.serverHost = socket.getfqdn("")
        self.asyncServer = None
        self.instanceClient = None
        self.guiConfiguration = None

        self.idCount = {}  # server only
        self.clients = {}  # server only

        # self.serverId = None # to remove
        # self.controlId = None # to remove
        # self.myProposal = None # to remove

        self.clientId2 = [None, None]  # client only
        self.serverId2 = [None, None]  # client AND server
        self.controlId2 = [None, None]  # client AND server

        self.bricksEventCache = {}

        global INSTANCE_HO
        INSTANCE_HO = self

        # Check the HO configuration
        if self.serverPort is None:
            self.log.error("InstanceServer: you must specify a port number")
        else:
            pass

    def initialize_instance(self):
        for widget in qt_import.QApplication.allWidgets():
            try:
                if hasattr(widget, "configuration"):
                    self.guiConfiguration = widget.configuration
                    break
            except NameError:
                logging.getLogger().warning(
                    "Widget {} has no attribute {}".format(widget, "configuration")
                )

        self.emit("instanceInitializing", ())
        if self.isLocal():
            self.startServer()
        else:
            self.connectToServer()

    def setProposal(self, proposal):
        if self.isServer():
            my_id = self.serverId2[0]
            old_proposal = self.serverId2[1]

            msg = NewClientInstanceMessage()
            msg.setClientId(my_id)
            msg.setAvailable(True)
            msg.setProposal(old_proposal)
            msg.setNewProposal(proposal)
            data = msg.encode()
            broadcast_to_clients(data)

            self.serverId2[1] = proposal
            if self.controlId2[0] == my_id:
                self.controlId2[1] = proposal
        elif self.isClient():
            my_id = self.clientId2[0]

            msg = NewClientInstanceMessage()
            msg.setNewProposal(proposal)
            data = msg.encode()
            send_data_to_server(self.instanceClient, data)

            self.clientId2[1] = proposal
            if self.controlId2[0] == my_id:
                self.controlId2[1] = proposal

    def idPrettyPrint(self, user_id, use_proposal=True):
        my_nick = ""
        pretty_print = ""
        if self.isServer():
            my_nick = self.serverId2[0]
        elif self.isClient():
            my_nick = self.clientId2[0]
        else:
            self.log.warning(
                "InstanceServer: printing an id while not server nor client"
            )

        if user_id is None:
            pretty_print = my_nick
        else:
            nick = user_id[0]
            prop = user_id[1]
            if my_nick != nick:
                if use_proposal:
                    if prop is not None:
                        pretty_print = "[%s%s]%s" % (prop["code"], prop["number"], nick)
                    else:
                        pretty_print = "[?]%s" % nick
                else:
                    pretty_print = nick
            else:
                pretty_print = my_nick

        return pretty_print

    # Starts the server
    def startServer(self):
        if self.asyncServer is not None:
            self.log.error("InstanceServer: server already started")
        elif self.serverPort is not None:
            try:
                async_server = gevent.server.StreamServer(
                    (self.serverHost, self.serverPort), handleRemoteClient
                )  # AsyncServer(self,self.serverHost,self.serverPort)
                async_server.start()
            except Exception:
                self.log.warning(
                    "InstanceServer: cannot create server, so trying to connect to it"
                )
                self.connectToServer()
            else:
                self.asyncServer = async_server
                server_hostname = self.serverHost.split(".")[0]
                self.log.debug(
                    "InstanceServer: listening to connections on %s:%d"
                    % (server_hostname, self.serverPort)
                )

                self.serverId2[0] = server_hostname
                self.controlId2 = list(self.serverId2)

                self.idCount[server_hostname] = 1

                self.emit("serverInitialized", (True, self.serverId2))
        else:
            self.log.error(
                "InstanceServer: not property configured to start the server"
            )
            self.emit("serverInitialized", (False,))

    # Closes the server
    def closeServer(self):
        self.asyncServer.close()
        self.asyncServer = None

    def connectToServer(self, quiet=False):
        self.emit("clientInitialized", (None,))
        self.reconnect(quiet)

    # Connects to the server
    def reconnect(self, quiet=False):
        try:
            self.instanceClient = InstanceClient(self.serverHost, self.serverPort)
        except Exception:
            self.instanceClient = None
            if not quiet:
                self.log.error("InstanceServer: cannot connect to server")
            self.emit("clientInitialized", (False, (None, None), None, quiet))
        else:
            my_login = pwd.getpwuid(os.getuid())[0]
            msg = AskPermissionInstanceMessage()

            msg.setClientId(my_login)
            msg.setProposal(self.clientId2[1])
            data = msg.encode()

            send_data_to_server(self.instanceClient, data)

    def isLocal(self):
        try:
            display = os.environ["DISPLAY"].split(":")[0]
        except Exception:
            return False
        if not display:
            return True
        return socket.getfqdn(display) == self.serverHost

    def isServer(self):
        return self.asyncServer is not None

    def isClient(self):
        return self.instanceClient is not None

    def inControl(self):
        return self.controlId2

    def serverClosed(self):
        self.log.error("InstanceServer: server has closed the connection!")
        self.emit("serverClosed", (self.serverId2,))

    def clientConnected(self, addr, req_handler):
        return

    def clientClosed(self, addr):
        # print "CLIENT CLOSED",addr,self.clients
        found_id = None
        found_prop = None
        for cli_id in self.clients:
            cli_addr = self.clients[cli_id][0]
            cli_prop = self.clients[cli_id][1]
            if cli_addr == addr:
                self.clients.pop(cli_id)
                found_id = cli_id
                found_prop = cli_prop
                self.emit("clientClosed", ((cli_id, cli_prop),))
                break

        if found_id is not None:
            msg = NewClientInstanceMessage()
            msg.setClientId(found_id)
            msg.setProposal(found_prop)
            msg.setAvailable(False)
            data = msg.encode()
            broadcast_to_clients(data)

            if self.controlId2[0] == found_id:
                self.controlId2 = list(self.serverId2)

                msg = PassControlInstanceMessage()
                msg.setClientId(self.controlId2[0])
                msg.setProposal(self.controlId2[1])
                data = msg.encode()
                broadcast_to_clients(data)

                self.emit("haveControl", (True,))

    def requestIdChange(self, new_id):
        if self.isServer():
            msg = NewClientInstanceMessage()

            try:
                count = self.idCount[new_id]
            except KeyError:
                self.idCount[new_id] = 1
            else:
                count += 1
                self.idCount[new_id] = count
                new_id = "%s-%d" % (new_id, count)
                self.idCount[new_id] = 1

            msg.setClientNewId(new_id)
            msg.setClientId(self.serverId2[0])
            msg.setAvailable(True)
            msg.setProposal(self.serverId2[1])
            data = msg.encode()
            broadcast_to_clients(data)

            old_id = self.serverId2[0]
            old_prop = self.serverId2[1]
            self.serverId2[0] = new_id
            if self.controlId2[0] == old_id:
                self.controlId2[0] = new_id
            new_prop = old_prop

            self.emit("clientChanged", ((old_id, old_prop), (new_id, new_prop)))

        elif self.isClient():
            msg = NewClientInstanceMessage()
            msg.setClientNewId(new_id)
            data = msg.encode()
            send_data_to_server(self.instanceClient, data)
        else:
            self.log.warning(
                "InstanceServer: requestIdChange while not server nor client!"
            )

    def sendChatMessage(self, priority, message):
        msg = ChatInstanceMessage()
        msg.setChatMessage(priority, message)
        if self.isServer():
            msg.setChatNick(self.serverId2[0])
            msg.setProposal(self.serverId2[1])
            my_id = self.serverId2
        data = msg.encode()
        if self.isServer():
            broadcast_to_clients(data)
        elif self.isClient():
            send_data_to_server(self.instanceClient, data)
            my_id = self.clientId2
        else:
            self.log.warning(
                "InstanceServer: sendChatMessage while not server nor client!"
            )
        self.emit("chatMessageReceived", (priority, my_id, message))

    def addEventToCache(self, brick_name, widget_name, message_data):
        self.bricksEventCache[(brick_name, widget_name)] = message_data

    def synchronizeClientWithEvents(self, client_addr):
        for event_data in list(self.bricksEventCache.values()):
            send_data_to_client(client_addr, event_data)

    def sendBrickUpdateMessage(
        self, brick_name, widget_name, widget_method, widget_method_args, masterSync
    ):
        msg = BrickUpdateInstanceMessage()
        msg.setBrickUpdate(
            brick_name, widget_name, widget_method, widget_method_args, masterSync
        )
        data = msg.encode()
        if self.isServer():
            self.addEventToCache(brick_name, widget_name, data)
            broadcast_to_clients(data)
        elif self.isClient():
            send_data_to_server(self.instanceClient, data)
        else:
            self.log.warning(
                "InstanceServer: sendBrickUpdateMessage while not server nor client!"
            )

    def sendTabUpdateMessage(self, tab_name, tab_index):
        msg = TabUpdateInstanceMessage()
        msg.setTabUpdate(tab_name, tab_index)
        data = msg.encode()
        self.addEventToCache(None, tab_name, data)
        if self.isServer():
            broadcast_to_clients(data)
        elif self.isClient():
            send_data_to_server(self.instanceClient, data)
        else:
            self.log.warning(
                "InstanceServer: sendTabUpdateMessage while not server nor client!"
            )

    def giveControl(self, client_id):
        if self.isServer():
            try:
                client_addr = self.clients[client_id[0]][0]
            except KeyError:
                pass
            else:
                client_prop = self.clients[client_id[0]][1]
                self.controlId2 = [client_id[0], client_prop]

                msg = ControlInstanceMessage()
                msg.setHasControl(True)
                data = msg.encode()
                send_data_to_client(client_addr, data)

                msg = PassControlInstanceMessage()
                msg.setClientId(client_id[0])
                msg.setProposal(client_prop)
                data = msg.encode()
                broadcast_to_clients(data, avoid=(client_addr,))

                self.emit("passControl", ((client_id[0], client_prop),))
                self.emit("haveControl", (False,))

        elif self.isClient():
            msg = PassControlInstanceMessage()
            msg.setClientId(client_id[0])
            msg.setProposal(client_id[1])
            data = msg.encode()
            send_data_to_server(self.instanceClient, data)
        else:
            self.log.warning("InstanceServer: giveControl while not server nor client!")

    def askForControl(self):
        msg = AskControlInstanceMessage()
        if self.isClient():
            data = msg.encode()
            send_data_to_server(self.instanceClient, data)
        else:
            msg.setClientId(self.serverId2[0])
            msg.setProposal(self.serverId2[1])
            data = msg.encode()
            broadcast_to_clients(data)

    def takeControl(self):
        if self.isServer():
            if self.controlId2[0] != self.serverId2[0]:
                client_addr = self.clients[self.controlId2[0]][0]
                msg = ControlInstanceMessage()
                msg.setHasControl(False)
                data = msg.encode()
                send_data_to_client(client_addr, data)

                self.controlId2 = list(self.serverId2)

                msg = PassControlInstanceMessage()
                msg.setClientId(self.controlId2[0])
                msg.setProposal(self.controlId2[1])
                data = msg.encode()
                broadcast_to_clients(data)

                self.emit("haveControl", (True,))
            else:
                self.log.warning(
                    "InstanceServer: takeControl while already in control!"
                )

        elif self.isClient():
            if self.controlId2[0] != self.clientId2[0]:
                msg = TakeControlInstanceMessage()
                data = msg.encode()
                send_data_to_server(self.instanceClient, data)
            else:
                self.log.warning(
                    "InstanceServer: takeControl while already in control!"
                )

    def callInControl(self, brick, method, method_args):
        msg = BrickCallInstanceMessage()
        brick_name = brick.name()
        widget_name = ""
        msg.setBrickUpdate(brick_name, widget_name, method, method_args)
        data = msg.encode()
        if self.isServer():
            if self.controlId2[0] != self.serverId2[0]:
                client_addr = self.clients[self.controlId2[0]][0]
                send_data_to_client(client_addr, data)
            else:
                self.log.warning(
                    "InstanceServer: calling a brick while having control!"
                )
        else:
            self.log.warning("InstanceServer: only the server can call a brick!")

    def answerToServer(self, brick, method, method_args):
        msg = BrickCallInstanceMessage()
        brick_name = brick.name()
        widget_name = ""
        msg.setBrickUpdate(brick_name, widget_name, method, method_args)
        data = msg.encode()
        if self.isServer():
            self.log.warning("InstanceServer: only a client can answer to the server!")
        else:
            data = msg.encode()
            send_data_to_server(self.instanceClient, data)

    def parseReceivedMessage(self, data):
        # logging.getLogger().debug("******** RECEIVED MESSAGE = %r", data)
        msg_obj = None
        try:
            message = InstanceMessage(data=data)
        except Exception:
            self.log.exception("InstanceServer: problem parsing received message")
        else:
            try:
                t = message.getType()
            except Exception:
                self.log.exception("InstanceServer: problem parsing received message")
            else:
                if t == InstanceMessage.TYPE_CHAT:
                    msg_obj = ChatInstanceMessage(message)
                elif t == InstanceMessage.TYPE_CONTROL:
                    msg_obj = ControlInstanceMessage(message)
                elif t == InstanceMessage.TYPE_ASKPERMISSION:
                    msg_obj = AskPermissionInstanceMessage(message)
                elif t == InstanceMessage.TYPE_GIVEPERMISSION:
                    msg_obj = GivePermissionInstanceMessage(message)
                elif t == InstanceMessage.TYPE_NEWCLIENT:
                    msg_obj = NewClientInstanceMessage(message)
                elif t == InstanceMessage.TYPE_ASKCONTROL:
                    msg_obj = AskControlInstanceMessage(message)
                elif t == InstanceMessage.TYPE_PASSCONTROL:
                    msg_obj = PassControlInstanceMessage(message)
                elif t == InstanceMessage.TYPE_BRICKUPDATE:
                    msg_obj = BrickUpdateInstanceMessage(message)
                elif t == InstanceMessage.TYPE_TABUPDATE:
                    msg_obj = TabUpdateInstanceMessage(message)
                elif t == InstanceMessage.TYPE_TAKECONTROL:
                    msg_obj = TakeControlInstanceMessage(message)
                elif t == InstanceMessage.TYPE_BRICKCALL:
                    msg_obj = BrickCallInstanceMessage(message)
                else:
                    self.log.warning(
                        "InstanceServer: unknown message type %s " % str(t)
                    )
        return msg_obj

    def clientMessageReceived(self, data):
        m = self.parseReceivedMessage(data)
        if m is None:
            return

        if isinstance(m, ChatInstanceMessage):
            self.emit(
                "chatMessageReceived",
                (
                    m.getChatPriority(),
                    (m.getChatNick(), m.getProposal()),
                    m.getChatMessage(),
                ),
            )
        elif isinstance(m, ControlInstanceMessage):
            has_control = m.getHasControl()
            self.emit("haveControl", (has_control,))

        elif isinstance(m, NewClientInstanceMessage):
            client_id = m.getClientId()
            client_proposal = m.getProposal()

            try:
                new_client_id = m.getClientNewId()
            except KeyError:
                try:
                    new_proposal = m.getNewProposal()
                except KeyError:
                    a = m.getAvailable()
                    if a:
                        self.emit("newClient", ((client_id, client_proposal),))
                    else:
                        self.emit("clientClosed", ((client_id, client_proposal),))
                else:
                    if client_id == self.serverId2[0]:
                        self.serverId2[1] = new_proposal
                    self.emit(
                        "clientChanged",
                        ((client_id, client_proposal), (client_id, new_proposal)),
                    )

            else:
                if client_id == self.serverId2[0]:
                    self.serverId2[0] = new_client_id
                self.emit(
                    "clientChanged",
                    ((client_id, client_proposal), (new_client_id, client_proposal)),
                )

        elif isinstance(m, GivePermissionInstanceMessage):
            self.serverId2 = [m.getServerId(), m.getProposal()]

            self.clientId2[0] = m.getClientId()

            self.emit(
                "clientInitialized",
                (True, (m.getServerId(), m.getProposal()), m.getClientId()),
            )

        elif isinstance(m, AskControlInstanceMessage):
            client_id = m.getClientId()
            try:
                client_prop = m.getProposal()
            except Exception:
                client_prop = None
            self.emit("wantsControl", ((client_id, client_prop),))

        elif isinstance(m, PassControlInstanceMessage):
            client_id = m.getClientId()
            client_proposal = m.getProposal()
            self.controlId2 = [client_id, client_proposal]
            self.emit("passControl", ((client_id, client_proposal),))

        elif isinstance(m, BrickCallInstanceMessage):
            try:
                timestamp = m.getTimestamp()
                brick_name = m.getBrickName()
                widget_name = m.getWidgetName()
                widget_method = m.getWidgetMethod()
                widget_method_args = m.getWidgetMethodArgs()
                brick = self.guiConfiguration.findItem(brick_name).brick

                if widget_name == "":
                    exec("method=brick.%s" % widget_method)
                else:
                    exec("method=brick.%s.%s" % (widget_name, widget_method))
                self.emit("widgetCall", (timestamp, method, widget_method_args))
            except Exception:
                self.log.exception("InstanceServer: problem while calling a brick!")

        elif isinstance(m, BrickUpdateInstanceMessage):
            try:
                timestamp = m.getTimestamp()
                brick_name = m.getBrickName()
                widget_name = m.getWidgetName()
                widget_method = m.getWidgetMethod()
                widget_method_args = m.getWidgetMethodArgs()
                masterSync = m.getMasterSync()

                # if necessary check this with Qt3 version
                brick = self.guiConfiguration.findItem(brick_name).get("brick")

                if widget_name == "":
                    exec("method=brick.%s" % widget_method)
                else:
                    exec("method=brick.%s.%s" % (widget_name, widget_method))
                self.emit(
                    "widgetUpdate", (timestamp, method, widget_method_args, masterSync)
                )
            except Exception:
                self.log.exception("InstanceServer: problem while updating a brick!")

        elif isinstance(m, TabUpdateInstanceMessage):
            try:
                timestamp = m.getTimestamp()
                tab_name = m.getTabName()
                tab_index = m.getTabIndex()
                tab = self.guiConfiguration.findItem(tab_name).widget

                method = tab.setCurrentPage
                method_args = (tab_index,)
                self.emit("widgetUpdate", (timestamp, method, method_args))
            except Exception:
                self.log.exception("InstanceServer: problem while updating a tab!")

    def serverMessageReceived(self, client_addr, data):
        m = self.parseReceivedMessage(data)
        if m is None:
            return

        if isinstance(m, NewClientInstanceMessage):
            try:
                client_new_id = m.getClientNewId()
            except KeyError:
                new_proposal = m.getNewProposal()
                # print "SERVER NEWCLIENT PROPOSAL",new_proposal,self.clients

                found_id = None
                for cli_id in self.clients:
                    cli_addr = self.clients[cli_id][0]
                    if cli_addr == client_addr:
                        found_id = cli_id
                        cli_prop = self.clients[cli_id][1]
                        msg = NewClientInstanceMessage()
                        msg.setClientId(cli_id)
                        msg.setProposal(cli_prop)
                        msg.setNewProposal(new_proposal)
                        msg.setAvailable(True)
                        data = msg.encode()
                        broadcast_to_clients(data)
                        self.emit(
                            "clientChanged",
                            ((cli_id, cli_prop), (cli_id, new_proposal)),
                        )
                        break

                if found_id is not None:
                    if self.controlId2[0] == found_id:
                        self.controlId2[1] = new_proposal
                    self.clients[cli_id][1] = new_proposal

            else:
                try:
                    count = self.idCount[client_new_id]
                except KeyError:
                    self.idCount[client_new_id] = 1
                else:
                    count += 1
                    self.idCount[client_new_id] = count
                    client_new_id = "%s-%d" % (client_new_id, count)
                    self.idCount[client_new_id] = 1

                found_id = None
                for cli_id in self.clients:
                    cli_addr = self.clients[cli_id][0]
                    if cli_addr == client_addr:
                        found_id = cli_id
                        cli_prop = self.clients[cli_id][1]
                        msg = NewClientInstanceMessage()
                        msg.setClientId(cli_id)
                        msg.setClientNewId(client_new_id)
                        msg.setProposal(cli_prop)
                        msg.setAvailable(True)
                        data = msg.encode()
                        broadcast_to_clients(data)
                        self.emit(
                            "clientChanged",
                            ((cli_id, cli_prop), (client_new_id, cli_prop)),
                        )
                        break

                if found_id is not None:
                    if self.controlId2[0] == found_id:
                        self.controlId2[0] = client_new_id
                    self.clients[client_new_id] = self.clients[found_id]
                    self.clients.pop(found_id)

        elif isinstance(m, AskPermissionInstanceMessage):
            client_id = m.getClientId()
            client_proposal = m.getProposal()

            try:
                count = self.idCount[client_id]
            except KeyError:
                self.idCount[client_id] = 1
            else:
                count += 1
                self.idCount[client_id] = count
                client_id = "%s-%d" % (client_id, count)
                self.idCount[client_id] = 1

            self.clients[client_id] = [client_addr, client_proposal]

            msg = NewClientInstanceMessage()
            msg.setClientId(client_id)
            msg.setProposal(client_proposal)
            msg.setAvailable(True)
            data = msg.encode()
            broadcast_to_clients(data, avoid=(client_addr,))

            for cli_id in self.clients:
                if cli_id != client_id:
                    msg = NewClientInstanceMessage()
                    cli_addr = self.clients[cli_id][0]
                    cli_prop = self.clients[cli_id][1]
                    msg.setClientId(cli_id)
                    msg.setProposal(cli_prop)
                    msg.setAvailable(True)
                    data = msg.encode()
                    send_data_to_client(client_addr, data)

            self.synchronizeClientWithEvents(client_addr)

            msg = GivePermissionInstanceMessage()
            msg.setClientId(client_id)
            msg.setServerId(self.serverId2[0])
            msg.setProposal(self.serverId2[1])
            data = msg.encode()
            send_data_to_client(client_addr, data)

            msg = PassControlInstanceMessage()
            msg.setClientId(self.controlId2[0])
            msg.setProposal(self.controlId2[1])
            data = msg.encode()
            send_data_to_client(client_addr, data)

            self.emit("newClient", ((client_id, client_proposal),))

        elif isinstance(m, ChatInstanceMessage):
            found_id = None

            for cli_id in self.clients:
                cli_addr = self.clients[cli_id][0]
                if cli_addr == client_addr:
                    found_id = cli_id
                    found_prop = self.clients[cli_id][1]
            if found_id is not None:
                m.setChatNick(found_id)
                m.setProposal(found_prop)
                data = m.encode()
                broadcast_to_clients(data, avoid=(client_addr,))
                self.emit(
                    "chatMessageReceived",
                    (m.getChatPriority(), (found_id, found_prop), m.getChatMessage()),
                )

        elif isinstance(m, AskControlInstanceMessage):
            found_id = None
            found_prop = None
            for cli_id in self.clients:
                cli_addr = self.clients[cli_id][0]
                if cli_addr == client_addr:
                    found_id = cli_id
                    found_prop = self.clients[cli_id][1]

            if found_id is not None:
                m.setClientId(found_id)
                if found_prop is not None:
                    m.setProposal(found_prop)
                data = m.encode()
                broadcast_to_clients(data)

                self.emit("wantsControl", ((found_id, found_prop),))

        elif isinstance(m, PassControlInstanceMessage):
            client_id = m.getClientId()
            if self.serverId2[0] == client_id:
                self.takeControl()
            else:
                try:
                    cli_addr = self.clients[client_id][0]
                except KeyError:
                    pass
                else:
                    client_prop = self.clients[client_id][1]

                    msg = ControlInstanceMessage()
                    msg.setHasControl(False)
                    data = msg.encode()
                    send_data_to_client(client_addr, data)

                    msg = PassControlInstanceMessage()
                    msg.setClientId(client_id)
                    msg.setProposal(client_prop)
                    data = msg.encode()
                    broadcast_to_clients(data, avoid=(cli_addr,))

                    self.controlId2 = [client_id, client_prop]
                    self.emit("passControl", ((client_id, client_prop),))

                    msg = ControlInstanceMessage()
                    msg.setHasControl(True)
                    data = msg.encode()
                    send_data_to_client(cli_addr, data)

        elif isinstance(m, BrickCallInstanceMessage):
            try:
                timestamp = m.getTimestamp()
                brick_name = m.getBrickName()
                widget_name = m.getWidgetName()
                widget_method = m.getWidgetMethod()
                widget_method_args = m.getWidgetMethodArgs()
                brick = self.guiConfiguration.findItem(brick_name).brick

                if widget_name == "":
                    exec("method=brick.%s" % widget_method)
                else:
                    exec("method=brick.%s.%s" % (widget_name, widget_method))
                self.emit("widgetCall", (timestamp, method, widget_method_args))
            except Exception:
                self.log.exception("InstanceServer: problem while calling a brick!")

        elif isinstance(m, BrickUpdateInstanceMessage):
            broadcast_to_clients(data, avoid=(client_addr,))

            try:
                timestamp = m.getTimestamp()
                brick_name = m.getBrickName()
                widget_name = m.getWidgetName()
                widget_method = m.getWidgetMethod()
                widget_method_args = m.getWidgetMethodArgs()
                brick = self.guiConfiguration.findItem(brick_name).brick
                masterSync = m.getMasterSync()
                if widget_name == "":
                    exec("method=brick.%s" % widget_method)
                else:
                    exec("method=brick.%s.%s" % (widget_name, widget_method))
                self.emit(
                    "widgetUpdate", (timestamp, method, widget_method_args, masterSync)
                )
            except Exception:
                self.log.exception("InstanceServer: problem while updating a brick!")

        elif isinstance(m, TabUpdateInstanceMessage):
            broadcast_to_clients(data, avoid=(client_addr,))

            try:
                timestamp = m.getTimestamp()
                tab_name = m.getTabName()
                tab_index = m.getTabIndex()
                tab = self.guiConfiguration.findItem(tab_name).widget

                method = tab.setCurrentPage
                method_args = (tab_index,)
                self.emit("widgetUpdate", (timestamp, method, method_args))
            except Exception:
                self.log.exception("InstanceServer: problem while updating a tab!")

        elif isinstance(m, TakeControlInstanceMessage):
            found_id = None
            found_prop = None
            for cli_id in self.clients:
                cli_addr = self.clients[cli_id][0]
                if cli_addr == client_addr:
                    found_id = cli_id
                    found_prop = self.clients[cli_id][1]

            if found_id is not None:
                if self.controlId2[0] == self.serverId2[0]:
                    self.emit("haveControl", (False,))
                else:
                    control_addr = self.clients[self.controlId2[0]][0]
                    msg = ControlInstanceMessage()
                    msg.setHasControl(False)
                    data = msg.encode()
                    send_data_to_client(control_addr, data)

                msg = PassControlInstanceMessage()
                msg.setClientId(found_id)
                msg.setProposal(found_prop)
                data = msg.encode()
                broadcast_to_clients(data)

                self.controlId2 = [found_id, found_prop]
                self.emit("passControl", ((found_id, found_prop),))

                msg = ControlInstanceMessage()
                msg.setHasControl(True)
                data = msg.encode()
                send_data_to_client(cli_addr, data)


def handleRemoteClient(client_socket, addr):
    SERVER_CLIENTS[addr] = client_socket
    INSTANCE_HO.clientConnected(addr, client_socket)

    buffer = ""
    while True:
        data = client_socket.recv(1024)
        if data == "":
            try:
                SERVER_CLIENTS.pop(addr)
            except Exception:
                # Huh? socket closed but client is not in SERVER_CLIENTS dict!
                # just ignore silently
                pass
            else:
                INSTANCE_HO.clientClosed(addr)
            break
        buffer += data
        msgs = buffer.split(TERMINATOR)
        buffer = msgs.pop()
        for msg in msgs:
            INSTANCE_HO.serverMessageReceived(addr, msg)


def broadcast_to_clients(data, avoid=None):
    for client_addr in list(SERVER_CLIENTS.keys()):
        if avoid and client_addr in avoid:
            continue
        send_data_to_client(client_addr, data)


def send_data_to_client(client_addr, data):
    client_socket = SERVER_CLIENTS.get(client_addr)
    if client_socket:
        try:
            client_socket.sendall("%s%s" % (data, TERMINATOR))
        except Exception:
            # broken pipe? client disconnected
            SERVER_CLIENTS.pop(client_addr)


def InstanceClient(host, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, port))
    except Exception:
        raise
    else:
        socketName = s.getsockname()

    def handle_incoming_data(client_socket):
        buffer = ""
        while True:
            data = client_socket.recv(1024)
            if data == "":
                INSTANCE_HO.serverClosed()
                break
            buffer += data
            msgs = buffer.split(TERMINATOR)
            buffer = msgs.pop()
            for msg in msgs:
                INSTANCE_HO.clientMessageReceived(msg)

    CLIENTS[socketName] = s
    gevent.spawn(handle_incoming_data, s)

    return socketName


def send_data_to_server(socket_name, data):
    client_socket = CLIENTS[socket_name]
    client_socket.sendall("%s%s" % (data, TERMINATOR))


class InstanceMessage:
    (
        TYPE_CHAT,
        TYPE_CONTROL,
        TYPE_NEWCLIENT,
        TYPE_ASKPERMISSION,
        TYPE_GIVEPERMISSION,
        TYPE_ASKCONTROL,
        TYPE_PASSCONTROL,
        TYPE_BRICKUPDATE,
        TYPE_TABUPDATE,
        TYPE_TAKECONTROL,
        TYPE_BRICKCALL,
    ) = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

    def __init__(self, data=None):
        self.messageDict = {}
        if data is not None:
            self.messageDict = pickle.loads(data)

    def encode(self):
        try:
            self.messageDict["type"]
        except KeyError:
            raise ValueError
        return pickle.dumps(self.messageDict)

    def getType(self):
        try:
            t = self.messageDict["type"]
        except KeyError:
            raise ValueError
        return t


class ChatInstanceMessage(InstanceMessage):
    (PRIORITY_LOW, PRIORITY_NORMAL, PRIORITY_HIGH) = (0, 1, 2)

    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_CHAT

    def setChatMessage(self, priority, message):
        self.messageDict["priority"] = priority
        self.messageDict["message"] = message

    def setChatNick(self, nick):
        self.messageDict["nick"] = nick

    def getChatMessage(self):
        return self.messageDict["message"]

    def getChatPriority(self):
        return self.messageDict["priority"]

    def getChatNick(self):
        return self.messageDict["nick"]

    def setProposal(self, proposal):
        self.messageDict["proposal"] = proposal

    def getProposal(self):
        return self.messageDict["proposal"]


class ControlInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_CONTROL

    def setHasControl(self, has_control):
        self.messageDict["control"] = has_control

    def getHasControl(self):
        return self.messageDict["control"]


class PassControlInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_PASSCONTROL

    def setClientId(self, client_id):
        self.messageDict["client_id"] = client_id

    def getClientId(self):
        return self.messageDict["client_id"]

    def setProposal(self, proposal):
        self.messageDict["proposal"] = proposal

    def getProposal(self):
        return self.messageDict["proposal"]


class AskControlInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_ASKCONTROL

    def setClientId(self, client_id):
        self.messageDict["client_id"] = client_id

    def getClientId(self):
        return self.messageDict["client_id"]

    def setProposal(self, proposal):
        self.messageDict["proposal"] = proposal

    def getProposal(self):
        return self.messageDict["proposal"]


class AskPermissionInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_ASKPERMISSION

    def setClientId(self, client_id):
        self.messageDict["client_id"] = client_id

    def getClientId(self):
        return self.messageDict["client_id"]

    def setProposal(self, proposal):
        self.messageDict["proposal"] = proposal

    def getProposal(self):
        return self.messageDict["proposal"]


class GivePermissionInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_GIVEPERMISSION

    def setClientId(self, client_id):
        self.messageDict["client_id"] = client_id

    def getClientId(self):
        return self.messageDict["client_id"]

    def setServerId(self, server_id):
        self.messageDict["server_id"] = server_id

    def getServerId(self):
        return self.messageDict["server_id"]

    def setProposal(self, proposal):
        self.messageDict["proposal"] = proposal

    def getProposal(self):
        return self.messageDict["proposal"]


class NewClientInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_NEWCLIENT

    def setClientId(self, client_id):
        self.messageDict["client_id"] = client_id

    def getClientId(self):
        return self.messageDict["client_id"]

    def setClientNewId(self, client_new_id):
        self.messageDict["client_new_id"] = client_new_id

    def getClientNewId(self):
        return self.messageDict["client_new_id"]

    def setAvailable(self, available):
        self.messageDict["available"] = available

    def getAvailable(self):
        return self.messageDict["available"]

    def setProposal(self, proposal):
        self.messageDict["proposal"] = proposal

    def getProposal(self):
        return self.messageDict["proposal"]

    def setNewProposal(self, proposal):
        self.messageDict["new_proposal"] = proposal

    def getNewProposal(self):
        return self.messageDict["new_proposal"]


class BrickUpdateInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_BRICKUPDATE

    def setBrickUpdate(
        self,
        brick_name,
        widget_name,
        widget_method,
        widget_method_args,
        masterSync=True,
    ):
        self.messageDict["timestamp"] = time.time()
        self.messageDict["brick_name"] = brick_name
        self.messageDict["widget_name"] = widget_name
        self.messageDict["widget_method"] = widget_method
        self.messageDict["widget_method_args"] = widget_method_args
        self.messageDict["masterSync"] = masterSync

    def getTimestamp(self):
        return self.messageDict["timestamp"]

    def getBrickName(self):
        return self.messageDict["brick_name"]

    def getWidgetName(self):
        return self.messageDict["widget_name"]

    def getWidgetMethod(self):
        return self.messageDict["widget_method"]

    def getWidgetMethodArgs(self):
        return self.messageDict["widget_method_args"]

    def getMasterSync(self):
        return self.messageDict["masterSync"]


class TabUpdateInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_TABUPDATE

    def setTabUpdate(self, tab_name, tab_index):
        self.messageDict["timestamp"] = time.time()
        self.messageDict["tab_name"] = tab_name
        self.messageDict["tab_index"] = tab_index

    def getTimestamp(self):
        return self.messageDict["timestamp"]

    def getTabName(self):
        return self.messageDict["tab_name"]

    def getTabIndex(self):
        return self.messageDict["tab_index"]


class TakeControlInstanceMessage(InstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_TAKECONTROL


class BrickCallInstanceMessage(BrickUpdateInstanceMessage):
    def __init__(self, instance_message=None):
        InstanceMessage.__init__(self)
        if instance_message is not None:
            self.messageDict = instance_message.messageDict
        self.messageDict["type"] = InstanceMessage.TYPE_BRICKCALL
